#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <sys/time.h>
#include "Client.h"

// number of one-second timeouts to give the server in asking for a response 
#define WAIT_TIME 10

Client::Client()
{
  connected = 0;
};

Client::Client(char *hostname, char *group, char *executable, int verbosity)
{
    assert (hostname && group && executable);
    connected = 0;
    sprintf(server_group, group);
    sprintf(server_hostname, hostname);
    reconnect(hostname, group, verbosity);
    if (!connected) {
	server_tid = spawnserver(hostname, executable, verbosity);
	if (server_tid > 0)
	    reconnect(server_tid, verbosity);
    }
};

int Client::disconnect() {
  int retval = -1;
  if (connected) {
    pvm_initsend(PvmDataDefault);
    retval = pvm_send(server_tid, DISCONNECTION_MSGTAG);  
  }; 
  return retval;
};
 
int Client::reconnect(char* hostname, char* group, int verbosity)
{
    int noserver = 0;
    char buf[100];
    int bufid = 0;   

    if (connected)
	disconnect();

    connected = 0;
    verbosity_level = verbosity;

    if ((!server_hostname && !hostname)
	|| (!server_group && !group)) {
	if (showErrors())
	    fprintf(stderr, "Missing info to establish connection.\n");
	return 0;
    };
  
    if (!hostname)
	hostname = server_hostname;
    if (!server_group)
	group = server_group;
  
    // save new information
    sprintf(server_group, group);
    sprintf(server_hostname, hostname);

    noserver = pvm_gsize(server_group);
  
    if (showTrace()) {
	fprintf(stderr,
		"Searching for match among %d registered servers ...\n",
		noserver);
    };
    if (noserver == PvmNoGroup) {
	if (showTrace())
	    fprintf(stderr,
		    "WARNING: No server found, client in unconnected state");
	return connected;
    }
    else {
	for (int servernum = 0; servernum < noserver; servernum++) {
	    server_tid = pvm_gettid(server_group, servernum);
	    if (server_tid < 0) { 
		if (showErrors())
		    fprintf(stderr, "Could not get tid for server!\n");
		return connected;
	    };
	    if (showTrace()) {
		fprintf(stderr, "Existing server found at tid %d.\n",
			server_tid);
		fprintf(stderr, "Attempting connection, ");
	    };
	    strcpy(buf, hostname);
	    pvm_initsend(PvmDataDefault);
	    pvm_pkstr(buf);
	    pvm_send(server_tid, CONNECTION_MSGTAG);
	    timeval tval;
	    tval.tv_sec = 1;
	    tval.tv_usec = 0;
	    int i;
	    for (i = 0; i < WAIT_TIME && bufid == 0; i++) { 
		bufid = pvm_trecv(server_tid, -1, &tval);	
		if (showTrace()) fprintf(stderr, ".");
	    };
	    if (bufid > 0) {
		if (showTrace())
		    fprintf(stderr, " connected!\n");
		pvm_unpackf("%d%s", &connected, buf);
	    } else {
		if (showTrace()) {
		    fprintf(stderr, " giving up.\n");
		    // Does this really do anything useful???
		    fprintf(stderr,
		            "Conection failed; message from server:\n\t%s\n",
		             buf);
		}
	    };
	    if (connected > 0)
		break;
	    else
		connected = 0;
	};
    };
    if (!connected)
	if (showTrace())
	    fprintf(stderr, "WARNING: Client in unconnected state.\n");
    return connected;
};

int Client::reconnect(int newservertid, int verbosity)
{
    char buf[64];

    if (connected)
	disconnect();

    connected = 0;
    verbosity_level = verbosity;
    server_tid = newservertid;

    if (showTrace()) {
	fprintf(stderr, "Attempting connection with server at tid %d, ");
    };
    strcpy(buf, "");
    pvm_initsend(PvmDataDefault);
    pvm_pkstr(buf);
    pvm_send(server_tid, CONNECTION_MSGTAG);
    timeval tval;
    tval.tv_sec = 1;
    tval.tv_usec = 0;
    int i, bufid = 0;
    for (i = 0; i < WAIT_TIME && bufid == 0; i++) { 
	bufid = pvm_trecv(server_tid, -1, &tval);	
	if (showTrace()) fprintf(stderr, ".");
    };
    if (bufid > 0) {
	if (showTrace()) fprintf(stderr, " connected!\n");
	pvm_unpackf("%d%s", &connected, buf);
    } else {
	if (showTrace()) {
	    fprintf(stderr, " giving up.\n");
	    // Does this really do anything useful???
	    fprintf(stderr,
	            "Conection failed; message from server:\n\t%s\n",
	            buf);
	}
    }
    if (!connected)
	if (showTrace())
	    fprintf(stderr, "WARNING: Client in unconnected state.\n");
    return connected;
};


int Client::spawnserver(char* hostname, char* executable, int verbosity) {
    int newservertid;
  
    assert (hostname && executable);
    int retval = pvm_spawn(executable, NULL, PvmTaskHost, hostname,
			   1, &newservertid);

    if (newservertid < 0 && verbosity >= VL_ERRORS) {
	if (newservertid == -7)
	    fprintf(stderr, "Could not find server executable at %s\n", 
		    executable);
	else
	    pvm_perror("Client::spawnserver");
    };
  
    return newservertid;
}

