#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include "PioneerClient.h"

PioneerClient::PioneerClient() : Client() {};

PioneerClient::PioneerClient(char *hostname, char *group,
                             char *executable, int verbosity)
    : Client (hostname, group, executable, verbosity)
{
};

PioneerClient::~PioneerClient()
{
};

int PioneerClient::vm(float v, float rv, float *xp, float *yp, float *thetap)
{ 
    int bufid, retval=0;
    float x, y, theta;

    if (showTrace()) fprintf(stderr, "vm: v = %g, rv = %g; ", v, rv);
    pvm_initsend(PvmDataDefault);
    pvm_packf("%f%f", v, rv);
    bufid = sendandreceive(SS_REQUEST_MSGTAG_VM);
    pvm_unpackf("%d%f%f%f", &retval, &x, &y, &theta);
    if (showTrace())
	fprintf(stderr,
		"returned: x = %g, y = %g, theta = %g.\n",
		x, y, theta);
    if (!retval && showErrors())
	fprintf(stderr, "vm failed.\n");
    if (xp)
	*xp = x;
    if (yp)
	*yp = y;
    if (thetap)
	*thetap = theta;
    return retval;
};

// You can get back the position with this call too.
int PioneerClient::diff_vm(float lwv, float rwv,
			   float *xp, float *yp, float *thetap)
{ 
    int bufid, retval=0;
    float x, y, theta;

    if (showTrace())
	fprintf(stderr, "diff_vm: lwv = %g, rwv = %g; ", lwv, rwv);
    pvm_initsend(PvmDataDefault);
    pvm_packf("%f%f", lwv, rwv);
    bufid = sendandreceive(SS_REQUEST_MSGTAG_DIFF_VM);
    pvm_unpackf("%d%f%f%f", &retval, &x, &y, &theta);
    if (showTrace())
	fprintf(stderr,
		"returned: x = %g, y = %g, theta = %g.\n",
		x, y, theta);
    if (!retval && showErrors())
	fprintf(stderr, "diff_vm failed.\n");
    if (xp)
	*xp = x;
    if (yp)
	*yp = y;
    if (thetap)
	*thetap = theta;
    return retval;
};

// Currently limping is non-functional, but call does return position.
int PioneerClient::lp(float *xp = NULL, float *yp = NULL, float *thetap = NULL)
{ 
    int bufid, retval=0;
    float x, y, theta;

    if (showTrace()) fprintf(stderr, "lp; ");
    pvm_initsend(PvmDataDefault);
    bufid = sendandreceive(SS_REQUEST_MSGTAG_LP);
    pvm_unpackf("%d%f%f%f", &retval, &x, &y, &theta);
    if (showTrace())
	fprintf(stderr,
		"returned: x = %g, y = %g, theta = %g.\n",
		x, y, theta);
    if (!retval && showErrors())
	fprintf(stderr, "lp failed.\n");
    if (xp)
	*xp = x;
    if (yp)
	*yp = y;
    if (thetap)
	*thetap = theta;
    return retval;
};

int PioneerClient::st(float *xp = NULL, float *yp = NULL, float *thetap = NULL)
{ 
    int bufid, retval=0;
    float x, y, theta;

    if (showTrace()) fprintf(stderr, "st; ");
    pvm_initsend(PvmDataDefault);
    bufid = sendandreceive(SS_REQUEST_MSGTAG_ST);
    pvm_unpackf("%d%f%f%f", &retval, &x, &y, &theta);
    if (showTrace())
	fprintf(stderr,
		"returned: x = %g, y = %g, theta = %g.\n",
		x, y, theta);
    if (!retval && showErrors())
	fprintf(stderr, "st failed.\n");
    if (xp)
	*xp = x;
    if (yp)
	*yp = y;
    if (thetap)
	*thetap = theta;
    return retval;
};

int PioneerClient::set_pos(float x, float y, float theta)
{ 
    int bufid, retval = 0;

    if (showTrace())
	fprintf(stderr, "set_position: x = %g, y = %g, theta = %g.\n",
		x, y, theta);
    pvm_initsend(PvmDataDefault);
    pvm_packf("%f%f%f", x, y, theta);
    bufid = sendandreceive(SS_REQUEST_MSGTAG_SET_POS);
    pvm_unpackf("%d", &retval);
    if (!retval && showErrors())
	fprintf(stderr, "set_position failed.\n");
    return retval;
};

int PioneerClient::query_pos(float *xp, float *yp, float *thetap)
{
    int bufid, retval=0;
    float x, y, theta;

    if (showTrace())
	fprintf(stderr, "query_position; ");
    pvm_initsend(PvmDataDefault);
    bufid = sendandreceive(SS_REQUEST_MSGTAG_QUERY_POS);  
    pvm_unpackf("%d%f%f%f", &retval, &x, &y, &theta);
    if (showTrace())
	fprintf(stderr,
		"returned: x = %g, y = %g, theta = %g.\n",
		x, y, theta);
    if (!retval && showErrors())
	fprintf(stderr, "query_position failed.\n");
    if (xp)
	*xp = x;
    if (yp)
	*yp = y;
    if (thetap)
	*thetap = theta;
    return retval;
};


// Not implemented on server side as far as I can tell.
//
// int PioneerClient::set_wheel_acceleration(int lefta, int righta = -1)
// {
//     int bufid, retval = 0;
// 
//     if (showTrace())
// 	   fprintf(stderr, "set_wheel_acceleration: lefta = %d, righta = %d\n",
// 		   lefta, righta);
//     pvm_initsend(PvmDataDefault);
//     pvm_packf("%d%d",lefta,righta);
//     bufid = sendandreceive(SS_REQUEST_MSGTAG_SET_WHEEL_ACC);
//     pvm_unpackf("%d", &retval);
//     if (!retval && showErrors())
// 	   fprintf(stderr, "set_wheel_acceleration failed.\n");
//     return retval;
// };
 

// Don't know the use of this one. Comments say "used in pr command only".
// Does not seem to be implemented on the server side anyway.
//
// int PioneerClient::set_wheel_velocity(int leftv, int rightv = -1)
// { 
//     int bufid, retval = 0;
// 
//     if (showTrace())
// 	   fprintf(stderr, "set_wheel_velocity: leftv = %d, rightv = %d\n",
// 		   leftv, rightv);
//     pvm_initsend(PvmDataDefault);
//     pvm_packf("%d%d",leftv,rightv);
//     bufid = sendandreceive(SS_REQUEST_MSGTAG_SET_WHEEL_VEL);
//     pvm_unpackf("%d", &retval);
//     if (!retval && showErrors())
// 	   fprintf(stderr, "set_wheel_velocity failed.\n");
//     return retval;
// };
 
// int PioneerClient::turn_front_sonar_on(int rate =15)
// {
//   int bufid, retval = 0;
//   pvm_initsend(PvmDataDefault);
//   pvm_packf("%d",rate);
//   bufid = sendandreceive(SS_REQUEST_MSGTAG_FRONT_SONAR);
//   pvm_unpackf("%d", &retval);
//   if (!retval)
//     cout << "turn_front_sonar_on failed" << endl;
//   return retval;
// };
// 
// int PioneerClient::turn_frontfive_sonar_on(int rate =15)
// {
//   int bufid, retval = 0;
//   pvm_initsend(PvmDataDefault);
//   pvm_packf("%d",rate);
//   bufid = sendandreceive(SS_REQUEST_MSGTAG_FRONT_FIVE_SONAR);
//   pvm_unpackf("%d", &retval);
//   if (!retval)
//     cout << "turn_frontfive_sonar_on failed" << endl;
//   return retval;
// };
// 
// int PioneerClient::turn_all_sonar_on(int rate = 15)
// {
//   int bufid, retval = 0;
//   pvm_initsend(PvmDataDefault);
//   pvm_packf("%d",rate);
//   bufid = sendandreceive(SS_REQUEST_MSGTAG_ALL_SONAR);
//   pvm_unpackf("%d", &retval);
//   if (!retval)
//     cout << "turn_all_sonar_on failed" << endl;
//   return retval;
// };
// 
// int PioneerClient::specify_sonar_order(int* sonar_firing_order, int rate = 15)
// {
//   int bufid, retval = 0;
//   pvm_initsend(PvmDataDefault);
//   pvm_pkint(sonar_firing_order,17,1);
//   pvm_packf("%d",rate);
//   bufid = sendandreceive(SS_REQUEST_MSGTAG_SONAR_ORDER);
//   pvm_unpackf("%d", &retval);
//   if (!retval)
//     cout << "specify_sonar_order failed" << endl;
//   return retval;
// };
// 
// int PioneerClient::turn_sonar_off()
// {
//   int bufid, retval = 0;
//   pvm_initsend(PvmDataDefault);
//   bufid = sendandreceive(SS_REQUEST_MSGTAG_SONAR_OFF);
//   pvm_unpackf("%d", &retval);
//   if (!retval)
//     cout << "turn_sonar_off failed" << endl;
//   return retval;
// };
// 
// int PioneerClient::get_sonar(float* sonar_vector)
// {
//   int bufid, retval = 0;
//   pvm_initsend(PvmDataDefault);
//   bufid = sendandreceive(SS_REQUEST_MSGTAG_GET_SONAR);
//   pvm_upkint(&retval,1,1);
//   if (!retval)
//     cout << "get_sonar failed" << endl;
//   else 
//     pvm_upkfloat(sonar_vector,16,1);
//   return retval;
// };
 
// int PioneerClient::get_bumpers(int* bumper_vector)
// {
//   int bufid, retval = 0;
//   pvm_initsend(PvmDataDefault);
//   bufid = sendandreceive(SS_REQUEST_MSGTAG_GET_BUMPERS);
//   pvm_upkint(&retval,1,1);
//   if (!retval)
//     cout << "get_bumpers failed" << endl;
//   else 
//     pvm_upkint(bumper_vector,6,1);
//   return retval;
// };
 
int PioneerClient::get_status(int what, int* status)
{ 
    int bufid, retval = 0;

    if (showTrace()) fprintf(stderr, "get_status: what = %d; ", what);
    pvm_initsend(PvmDataDefault);
    pvm_packf("%d", what);
    bufid = sendandreceive(SS_REQUEST_MSGTAG_GET_STATUS);
    pvm_upkint(&retval,1,1);
    if (!retval) {
	if (showErrors()) fprintf(stderr, "get_status failed.\n");
    } else {
	pvm_upkint(status,1,1);
	if (showTrace()) fprintf(stderr, "returned: status = %d.\n", *status);
    }
    return retval;
};

int PioneerClient::get_battery_status(int* status)
{ 
    int bufid, retval = 0;

    if (showTrace()) fprintf(stderr, "get_battery_status; ");
    pvm_initsend(PvmDataDefault);
    bufid = sendandreceive(SS_REQUEST_MSGTAG_GET_BATTERY_STATUS);
    pvm_upkint(&retval,1,1);
    if (!retval) {
	if (showErrors()) fprintf(stderr, "get_battery_status failed.\n");
    } else {
	pvm_upkint(status,1,1);
	if (showTrace()) fprintf(stderr, "returned: status = %d.\n", *status);
    }
    return retval;
};
